/*
 * BPEXAMPLE : example of learning the exor problem.
 *
 * Made by : M. Kraaijveld
 *           W. Schmidt
 * SccsId  : %Z%$RCSfile: bpexam.c,v $ V%Q% $Revision: 1.2 $ $Date: 1993/10/07 13:55:05 $
 */

#include <math.h>
#include <stdio.h>
#include <sprlib.h>

extern void exit();

main(argc, argv)
int argc;
char *argv[];
{
	NET *net;
	SAMPLE *sample_ptr;
	DATASET *learning_set;
	FILE *stream;
	ERROR_STRUCT error_struct, *error_struct_ptr;
	int *NetVect;
	long options = 0, done = FALSE, cycles = 0, total_cycles = 0;

	/*
	 * Handle program options and usage.
	 */

	if (argc != 2)
	{
		fprintf(stderr, "usage: bpexample nr_cycles\n");
		exit(1);
	}
	cycles = atoi(argv[1]);


	/*
	 * Initialise the environment.
	 * Exit on error so return codes can be ignored.
	 */

	sprinit(INIT_EXIT);

	/*
	 * Load learning set, show it on the screen and initialise (LONG)
	 * performance computation
	 */

	learning_set = load_dataset("par2.dat");
	fprintf(stdout, "\n\nData Set:\n");
	fprintf_dataset(stdout, learning_set);
	perf_init(learning_set);

	/*
	 * Make Network, randomize weights and initialise backprop learning
	 */

	NetVect = ivector(1, 3);
	NetVect[1] = 2;
	NetVect[2] = 4;
	NetVect[3] = 1;
	net = create_ff_net(1L, 3, NetVect);
	unif_rand_net(1, -.01, 0.01, net);
	bp_init(net);

	/*
	 * Compute and print the performance of the network
	 */

	fprintf(stdout, "\n\nLearning Process:\n");
	if ((error_struct_ptr =
	     net_perf(net, learning_set, &error_struct, 0.1, LONG_PERF)) == NULL)
	{
		sprerror("Problem in computation of net performance.");
		exit(1);
	}

	fprintf(stdout, "cycle: %d, ", total_cycles);
	fprintf_error_struct(stdout, error_struct_ptr);

	/*
	 * Train the network with the backpropagation algorithm
	 * Options are: do not store the history of the units,
	 * the thresholds and the weights, and use the default
	 * backprop options: accumulate and update.
	 */

	options = (NO_HISTU | NO_HISTT | NO_HISTW | BPACCUM | BPUPDATE);

	while (done == FALSE)
	{

		/*
		 * Perform 'cycles' learning cycles, with eta = 1, alpha = 0.9
		 */

		if (bp_learn(net, learning_set, 1.0, 0.9, cycles, options))
		{
			sprerror("Problem in training network.");
			exit(1);
		}

		total_cycles += cycles;

		/*
		 * Store the weights and theta's in the history list
		 */

		if (store_hist_all_units(net))
		{
			sprerror("Problem in storing weight history.");
			exit(1);
		}

		if (store_hist_all_thetas(net))
		{
			sprerror("Problem in storing theta history.");
			exit(1);
		}

		if (store_hist_all_weights(net))
		{
			sprerror("Problem in storing weight history.");
			exit(1);
		}

		/*
		 * Compute and print the performance of the network, use the
		 * LONG_PERF option to store the error structures per sample
		 * in the SystemData union of the samples.
		 * Error threshold is 0.1
		 */

		if ((error_struct_ptr =
		     net_perf(net, learning_set, &error_struct, 0.1, LONG_PERF)) == NULL)
		{
			sprerror("Problem in computation of net performance.");
			exit(1);
		}

		fprintf(stdout, "cycle: %d, ", total_cycles);
		fprintf_error_struct(stdout, error_struct_ptr);

		/*
		 * Stop criteria
		 */
		if (error_struct.AllCorrect == TRUE)         done = TRUE;
		if (error_struct.PercentageCorrect > 99.0)   done = TRUE;
		if (error_struct.MeanSquaredError  < 1.0e-5) done = TRUE;
		if (total_cycles > 1000000L)                 done = TRUE;
	}

	printf("Memory usage %u\n", memory_monitor());

	/*
	 * Print the total status of the network on the screen
	 */

	fprintf(stdout, "\n\nNetwork Status:\n");
	status_total(stdout, net);

	/*
	 * Print the total performance of the learning set on the screen
	 */

	fprintf(stdout, "\n\nLearning Set Performance:\n");
	sample_ptr = learning_set->FirstSample;
	while (sample_ptr)
	{

		fprintf(stdout, "sample: %d, ", sample_ptr->SampleId);
		fprintf_error_struct(stdout, (ERROR_STRUCT *) sample_ptr->SystemData.p);

		sample_ptr = sample_ptr->Next;
	}
	fprintf(stdout, "\nTOTAL:     ", total_cycles);
	fprintf_error_struct(stdout, error_struct_ptr);

	/*
	 * Save the network (including its history) in a file and 
	 * free / delete network and learningset
	 */

	if ((stream = fopen("exor.net", "w")) == NULL)
	{
		sprerror("Cannot open file 'exor.net'.");
		exit(1);
	}
	fprintf_network(stream, net, (HISTU | HISTT | HISTW));
	fclose(stream);

	bp_free(net);
	delete_net(net);
	perf_free(learning_set);
	delete_dataset(learning_set);

	/*
	 * Free all structures and quit.
	 */

	free_ivector(NetVect, 1, 3);

	sprexit(EXIT_SILENT);

	return 0;
}

/*
 * End of Source
 */
